##
# File: ChemRefDataPrep.py
# Author:  J. Westbrook
# Date:  4-Dec-2018
#
# Very preliminary implementation for building loadable corresponding data from
# external resources (e.g. DrugBank).
#
# Update:
#
##

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import logging

from rcsb.utils.chemref.DrugBankUtils import DrugBankUtils
# from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class ChemRefDataPrep(object):
    """
    Integrate external annotations with local chemical reference data.

    """

    def __init__(self, cfgOb, **kwargs):
        #
        self.__cfgOb = cfgOb
        self.__resourceName = "MONGO_DB"
        # self.__workPath = kwargs.get('workPath', None)
        # self.__sandboxPath = kwargs.get('sandboxPath', None)
        #
        #self.__mU = MarshalUtil(workPath=self.__workPath)
        #

    def getDocuments(self, extResource, exIdD):
        """
        """
        oL = []
        if extResource == 'DrugBank':
            drugBankFilePath = self.__cfgOb.getPath('DRUGBANK_DATA_LOCATOR')
            dbDocL = self.fetchDocuments(extResource, drugBankFilePath)
            oL = []
            for dbDoc in dbDocL:
                if 'drugbank_id' in dbDoc and dbDoc['drugbank_id'] in exIdD:
                    lD = self.buildDocument(extResource, dbDoc)
                    oL.append(lD)
                else:
                    logger.debug("dbDoc.keys() %r" % list(dbDoc.keys()))
        return oL
        #

    def fetchDocuments(self, extResource, filePath):
        if extResource == "DrugBank":
            return self.__fetchDrugBankDocuments(filePath)
        else:
            return None

    def buildDocument(self, extResource, dbObj):
        if extResource == "DrugBank":
            return self.__buildDrugBankDocument(dbObj)
        else:
            return None

    def __fetchDrugBankDocuments(self, drugBankFilePath):
        """
        """
        dbu = DrugBankUtils()
        rL = dbu.read(drugBankFilePath)
        logger.info("DrugBank data object length %d" % len(rL))
        # logger.info("DrugBank example keys %r" % rL[0].keys())
        # logger.info("DrugBank example aliases %r" % rL[0]['aliases'])
        return rL

    def __buildDrugBankDocument(self, dbObj):
        """
        Construct local loadable object from input DrugBank extracted data object
        conforming to the following category organization:

         _drugbank_info.drugbank_id
         _drugbank_info.name
         _drugbank_info.description
         _drugbank_info.synonyms
         _drugbank_info.brand_names
         _drugbank_info.affected_organisms
         _drugbank_info.indication
         _drugbank_info.pharmacology
         _drugbank_info.mechanism_of_action
         _drugbank_info.cas_number
         _drugbank_info.drug_categories
         _drugbank_info.drug_groups


         _drugbank_target.ordinal
         _drugbank_target.name
         _drugbank_target.interaction_type
         _drugbank_target.target_actions
         _drugbank_target.organism_common_name
         _drugbank_target.reference_database_name
         _drugbank_target.reference_database_accession_code
         _drugbank_target.seq_one_letter_code

        """
        oD = {}
        oD['__drugbank_id'] = dbObj['drugbank_id']
        dbiD = {}
        textKeys = [('drugbank_id', 'drugbank_id'),
                    ('name', 'name'),
                    ('description', 'description'),
                    ('indication', 'indication'),
                    ('pharmacology', 'pharmacology'),
                    ('mechanism_of_action', 'mechanism_of_action'),
                    ('cas_number', 'cas_number')]
        listKeys = [('drug_categories', 'drug_categories'),
                    ('groups', 'drug_groups'),
                    ('aliases', 'synonyms'),
                    ('products', 'brand_names'),
                    ('affected-organisms', 'affected-organisms')]
        # For category drugbank_info
        for textKey, docKey in textKeys:
            if textKey in dbObj and len(dbObj[textKey]):
                dbiD[docKey] = dbObj[textKey].replace('\r', '').replace('\n', ' ')
        #
        for listKey, docKey in listKeys:
            if listKey in dbObj and len(dbObj[listKey]):
                dbiD[docKey] = list(dbObj[listKey])
            else:
                logger.debug("MISSING KEY %s" % listKey)
        #
        oD['drugbank_info'] = dbiD
        #
        # For category drugbank_target -
        #
        tL = []
        if 'target_interactions' in dbObj:
            for ii, tid in enumerate(dbObj['target_interactions'], 1):
                tD = {}
                tD['ordinal'] = ii
                tD['interaction_type'] = tid['category']
                tD['name'] = tid['name']
                if 'organism' in tid:
                    tD['organism_common_name'] = tid['organism']
                if 'actions' in tid and len(tid['actions']):
                    tD['target_actions'] = tid['actions']
                if 'amino-acid-sequence' in tid and tid['amino-acid-sequence'] and len(tid['amino-acid-sequence']) > 0:
                    tD['seq_one_letter_code'] = "".join(tid['amino-acid-sequence'].split('\n')[1:])
                if 'uniprot_ids' in tid and len(tid['uniprot_ids']):
                    tD['reference_database_name'] = 'UniProt'
                    tD['reference_database_accession_code'] = tid['uniprot_ids']
                #
                tL.append(tD)
        #
        if len(tL):
            oD['drugbank_target'] = tL
        #
        return oD



##
# File:    ChemCompModelUtils.py
# Author:  J. Westbrook
# Date:    1-Nov-2018
#
# Updates:

##
"""
Utilities to read resource file containing compilation of CCDC models correspondences
for PDB chemical components.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging

from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class ChemCompModelUtils(object):
    """
    """

    def __init__(self):
        self.__ns = None

    def __readCifFile(self, filePath):
        """ Read input CIF file and return a container list.
        """
        cL = []
        try:
            mU = MarshalUtil()
            cL = mU.doImport(filePath, format="mmcif")
            logger.debug("Container list %d" % len(cL))
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return cL

    def getMapping(self, modelFilePath):
        """

        Example:

        data_M_010_00001
            #
            _pdbx_chem_comp_model.id        M_010_00001
            _pdbx_chem_comp_model.comp_id   010
            #
            _pdbx_chem_comp_model_reference.model_id   M_010_00001
            _pdbx_chem_comp_model_reference.db_name    CSD
            _pdbx_chem_comp_model_reference.db_code    YARXEW
            #
            loop_
            _pdbx_chem_comp_model_feature.model_id
            _pdbx_chem_comp_model_feature.feature_name
            _pdbx_chem_comp_model_feature.feature_value
            M_010_00001 experiment_temperature 123.0
            M_010_00001 publication_doi        10.1021/cg200971f
            M_010_00001 r_factor               3.91
            M_010_00001 all_atoms_have_sites   Y
        #
        """
        rD = {}
        dataContainerL = self.__readCifFile(modelFilePath)

        for dataContainer in dataContainerL:
            logger.debug("Processing model %r" % dataContainer.getName())
            cObj = dataContainer.getObj('pdbx_chem_comp_model')
            ccId = cObj.getValue('comp_id', 0)
            #
            if ccId not in rD:
                rD[ccId] = []

            tObj = dataContainer.getObj('pdbx_chem_comp_model_reference')
            for iRow in range(tObj.getRowCount()):
                dbName = tObj.getValue('db_name', iRow)
                dbCode = tObj.getValue('db_code', iRow)
                rD[ccId].append({'db_name': dbName, 'db_code': dbCode})

        return rD



##
# File:    DrugBankUtils.py
# Author:  J. Westbrook
# Date:    17-Oct-2018
# Version: 0.001
#
# Update:
#
##
"""
Various utilities for extracting data from DrugBank repository data.

Approach is inspired by a useful examples at:

https://github.com/dhimmel/drugbank/blob/gh-pages/parse.ipynb
and https://github.com/bio2bel

"""
import gzip
import itertools as itt
import logging
import time
from datetime import datetime

try:
    import xml.etree.cElementTree as ET
except ImportError:
    import xml.etree.ElementTree as ET

logger = logging.getLogger(__name__)


class DrugBankUtils(object):
    """
    """

    def __init__(self):
        self.__ns = '{http://www.drugbank.ca}'

    def read(self, filePath):
        xrt = self.__parse(filePath)
        rL = []
        for drugElement in xrt:
            rL.append(self.__processDrugElement(drugElement))
        return rL
        #

    def __parse(self, filePath):
        """ Parse the iput DrugBank export XML data file.
        """
        tree = []
        if filePath[-3:] == '.gz':
            with gzip.open(filePath, mode='rb') as ifh:
                logger.debug('Parsing drugbank at %s', filePath)
                t = time.time()
                tree = ET.parse(ifh)
                logger.debug('Parsed drugbank in %.2f seconds', time.time() - t)
        else:
            with open(filePath, mode='rb') as ifh:
                logger.debug('Parsing drugbank at %s', filePath)
                t = time.time()
                tree = ET.parse(ifh)
                logger.debug('Parsed drugbank in %.2f seconds', time.time() - t)
        return tree.getroot()

    def __processDrugElement(self, drugElement):
        """
        """
        assert drugElement.tag == "{ns}drug".format(ns=self.__ns)

        doc = {
            'type': drugElement.get('type'),
            'drugbank_id': drugElement.findtext("{ns}drugbank-id[@primary='true']".format(ns=self.__ns)),
            'cas_number': drugElement.findtext("{ns}cas-number".format(ns=self.__ns)),
            'name': drugElement.findtext("{ns}name".format(ns=self.__ns)),
            'description': drugElement.findtext("{ns}description".format(ns=self.__ns)),
            'indication': drugElement.findtext("{ns}indication".format(ns=self.__ns)),
            'pharmacodynamics': drugElement.findtext("{ns}pharmacodynamics".format(ns=self.__ns)),
            'mechanism-of-action': drugElement.findtext("{ns}mechanism-of-action".format(ns=self.__ns)),
            'groups': [
                group.text
                for group in drugElement.findall("{ns}groups/{ns}group".format(ns=self.__ns))
            ],
            'affected-organisms': [
                org.text
                for org in drugElement.findall("{ns}affected-organisms/{ns}affected-organism".format(ns=self.__ns))
            ],
            'atc_codes': [
                code.get('code')
                for code in drugElement.findall("{ns}atc-codes/{ns}atc-code".format(ns=self.__ns))
            ],
            'categories': [
                {
                    'name': el.findtext("{ns}category".format(ns=self.__ns)),
                    'mesh_id': el.findtext('{ns}mesh-id'.format(ns=self.__ns))
                }
                for el in drugElement.findall("{ns}categories/{ns}category".format(ns=self.__ns))
            ],
            'patents': [
                {
                    'patent_id': el.findtext('{ns}number'.format(ns=self.__ns)),
                    'country': el.findtext('{ns}country'.format(ns=self.__ns)),
                    'approved': datetime.strptime(el.findtext('{ns}approved'.format(ns=self.__ns)), '%Y-%m-%d'),
                    'expires': datetime.strptime(el.findtext('{ns}expires'.format(ns=self.__ns)), '%Y-%m-%d'),
                    'pediatric_extension': el.findtext('{ns}pediatric-extension'.format(ns=self.__ns)) != 'false',

                }
                for el in drugElement.findall("{ns}patents/{ns}patent".format(ns=self.__ns))
            ],
            'external_identifiers': [
                {
                    'resource': el.findtext('{ns}resource'.format(ns=self.__ns)),
                    'identifier': el.findtext('{ns}identifier'.format(ns=self.__ns))
                }
                for el in drugElement.findall("{ns}external-identifiers/{ns}external-identifier".format(ns=self.__ns))
            ],
            'pdb_entries': [
                pdbId.text
                for pdbId in drugElement.findall("{ns}pdb-entries/{ns}pdb-entry".format(ns=self.__ns))
            ],
            'inchi': drugElement.findtext("{ns}calculated-properties/{ns}property[{ns}kind='InChI']/{ns}value".format(ns=self.__ns)),
            'inchikey': drugElement.findtext("{ns}calculated-properties/{ns}property[{ns}kind='InChIKey']/{ns}value".format(ns=self.__ns)),
            'exp_prop': [
                {
                    'kind': el.findtext('{ns}kind'.format(ns=self.__ns)),
                    'value': el.findtext('{ns}value'.format(ns=self.__ns)),
                    'source': el.findtext('{ns}source'.format(ns=self.__ns))
                }
                for el in drugElement.findall("{ns}experimental-properties/{ns}property".format(ns=self.__ns))
            ],
            'calc_prop': [
                {
                    'kind': el.findtext('{ns}kind'.format(ns=self.__ns)),
                    'value': el.findtext('{ns}value'.format(ns=self.__ns)),
                    'source': el.findtext('{ns}source'.format(ns=self.__ns))
                }
                for el in drugElement.findall("{ns}calculated-properties/{ns}property".format(ns=self.__ns))
            ],
            'products': [
                {
                    'name': el.findtext('{ns}name'.format(ns=self.__ns)),
                    'labeller': el.findtext('{ns}labeller'.format(ns=self.__ns)),
                    'ndc-id': el.findtext('{ns}ndc-id'.format(ns=self.__ns)),
                    'ndc-product-code': el.findtext('{ns}ndc-product-code'.format(ns=self.__ns)),
                    'dpd-id': el.findtext('{ns}dpd-id'.format(ns=self.__ns)),
                    'ema-product-code': el.findtext('{ns}ema-product-code'.format(ns=self.__ns)),
                    'ema-ma-number': el.findtext('{ns}ema-ma-number'.format(ns=self.__ns)),
                    'started-marketing-on': el.findtext('{ns}started-marketing-on'.format(ns=self.__ns)),
                    'ended-marketing-on': el.findtext('{ns}ended-marketing-on'.format(ns=self.__ns)),
                    'dosage-form': el.findtext('{ns}dosage-form'.format(ns=self.__ns)),
                    'strength': el.findtext('{ns}strength'.format(ns=self.__ns)),
                    'route': el.findtext('{ns}route'.format(ns=self.__ns)),
                    'fda-application-number': el.findtext('{ns}fda-application-number'.format(ns=self.__ns)),
                    'generic': el.findtext('{ns}generic'.format(ns=self.__ns)),
                    'over-the-counter': el.findtext('{ns}over-the-counter'.format(ns=self.__ns)),
                    'approved': el.findtext('{ns}approved'.format(ns=self.__ns)),
                    'country': el.findtext('{ns}country'.format(ns=self.__ns)),
                    'source': el.findtext('{ns}source'.format(ns=self.__ns))
                }
                for el in drugElement.findall("{ns}products/{ns}product".format(ns=self.__ns))
            ]
        }
        aliases = {
            elem.text.strip().encode('ascii', 'xmlcharrefreplace').decode("utf-8") for elem in
            itt.chain(
                drugElement.findall("{ns}synonyms/{ns}synonym".format(ns=self.__ns)),
                drugElement.findall("{ns}salts/{ns}salt/{ns}name".format(ns=self.__ns)),
            )
            if elem.text.strip()
        }
        aliases.add(doc['name'])
        doc['aliases'] = aliases

        products = {
            elem.text.strip().encode('ascii', 'xmlcharrefreplace').decode("utf-8") for elem in
            itt.chain(
                drugElement.findall("{ns}international-brands/{ns}international-brand".format(ns=self.__ns)),
                drugElement.findall("{ns}products/{ns}product/{ns}name".format(ns=self.__ns)),
            )
            if elem.text.strip()
        }
        doc['products'] = list(products)
        #
        doc['target_interactions'] = []
        targetCategories = ['target', 'enzyme', 'carrier', 'transporter']
        for category in targetCategories:
            targets = drugElement.findall('{ns}{category}s/{ns}{category}'.format(ns=self.__ns, category=category))
            for target in targets:
                targetDoc = self.__getTargetInfo(category, target)
                if not targetDoc:
                    continue
                doc['target_interactions'].append(targetDoc)

        if 'categories' in doc and len(doc['categories']):
            doc['drug_categories'] = [c['name'] for c in doc['categories']]
        return doc

    def __getTargetInfo(self, category, target):
        doc = {
            'category': category,
            'organism': target.findtext('{ns}organism'.format(ns=self.__ns)),
            'known_action': target.findtext('{ns}known-action'.format(ns=self.__ns)),
            'name': target.findtext('{ns}name'.format(ns=self.__ns)),
            'amino-acid-sequence': target.findtext('{ns}polypeptide/{ns}amino-acid-sequence[@format="FASTA"]'.format(ns=self.__ns)),
            'actions': [
                action.text
                for action in target.findall('{ns}actions/{ns}action'.format(ns=self.__ns))
            ],
            'articles': [
                pubmed_element.text
                for pubmed_element in target.findall('{ns}references/{ns}articles/{ns}article/{ns}pubmed-id'.format(ns=self.__ns))
                if pubmed_element.text
            ]
        }

        uniprot_ids = [polypep.text for polypep in target.findall(
            "{ns}polypeptide/{ns}external-identifiers/{ns}external-identifier[{ns}resource='UniProtKB']/{ns}identifier".format(ns=self.__ns))]
        if len(uniprot_ids) > 0:
            doc['uniprot_ids'] = uniprot_ids[0]

        hgnc_ids = [polypep.text for polypep in target.findall(
            "{ns}polypeptide/{ns}external-identifiers/{ns}external-identifier[{ns}resource='HUGO Gene Nomenclature Committee (HGNC)']/{ns}identifier".format(ns=self.__ns))]
        if len(hgnc_ids) == 1:
            doc['hgnc_id'] = hgnc_ids[0][len('HGNC:'):]

        return doc
